try:
    import pydantic  # type: ignore
except ImportError as e:
    raise ImportError(
        "camel-converter must be installed with the pydantic extra to use this class"
    ) from e

from camel_converter import to_camel


class CamelBase(pydantic.BaseModel):
    """A Pydantic model that provides a base configuration for converting between camel and snake case.

    If another Pydantic model inherit from this class it will get the ability to do this conversion
    between camel and snake case without having to add the configuration to the new model.
    """

    model_config = pydantic.ConfigDict(alias_generator=to_camel, populate_by_name=True)  # type: ignore[attr-defined]



from dataclasses import dataclass

import pytest

from camel_converter import (
    Converter,
    dict_to_camel,
    dict_to_pascal,
    dict_to_snake,
    to_camel,
    to_pascal,
    to_snake,
)


@pytest.mark.parametrize(
    "test_dict, expected",
    [
        (
            {
                "convert_me": "val 1",
                "me_also": "val 2",
                "unchanged": "val 3",
                1: "val 4",
            },
            {"convertMe": "val 1", "meAlso": "val 2", "unchanged": "val 3", 1: "val 4"},
        ),
        (
            {
                "convert_me": "val 1",
                "me_also": {
                    "nested_convert": 1,
                    "unchanged": 2,
                    "another_nest": {"one_more": "hi"},
                },
                1: {"another_nested": "a"},
            },
            {
                "convertMe": "val 1",
                "meAlso": {
                    "nestedConvert": 1,
                    "unchanged": 2,
                    "anotherNest": {"oneMore": "hi"},
                },
                1: {"anotherNested": "a"},
            },
        ),
        (
            {"test_convert": [{"change_me": 1, "unchanged": 2}, 1]},
            {"testConvert": [{"changeMe": 1, "unchanged": 2}, 1]},
        ),
        (
            {"test_convert": ({"change_me": 1, "unchanged": 2}, 1)},
            {"testConvert": ({"changeMe": 1, "unchanged": 2}, 1)},
        ),
    ],
)
def test_dict_to_camel(test_dict, expected):
    assert dict_to_camel(test_dict) == expected


@pytest.mark.parametrize(
    "test_dict, expected",
    [
        (
            {"convert_me": "val 1", "me_also": "val 2", "changed": "val 3", 1: "val 4"},
            {"ConvertMe": "val 1", "MeAlso": "val 2", "Changed": "val 3", 1: "val 4"},
        ),
        (
            {
                "convert_me": "val 1",
                "me_also": {
                    "nested_convert": 1,
                    "changed": 2,
                    "another_nest": {"one_more": "hi"},
                },
                1: {"another_nested": "a"},
            },
            {
                "ConvertMe": "val 1",
                "MeAlso": {
                    "NestedConvert": 1,
                    "Changed": 2,
                    "AnotherNest": {"OneMore": "hi"},
                },
                1: {"AnotherNested": "a"},
            },
        ),
        (
            {"test_convert": [{"change_me": 1, "changed": 2}, 1]},
            {"TestConvert": [{"ChangeMe": 1, "Changed": 2}, 1]},
        ),
        (
            {"test_convert": ({"change_me": 1, "changed": 2}, 1)},
            {"TestConvert": ({"ChangeMe": 1, "Changed": 2}, 1)},
        ),
    ],
)
def test_dict_to_pascal(test_dict, expected):
    assert dict_to_pascal(test_dict) == expected


@pytest.mark.parametrize(
    "test_dict, expected, treat_digits_as_capitals",
    [
        (
            {"convertMe": "val 1", "meAlso": "val 2", "unchanged": "val 3", 1: "val 4"},
            {
                "convert_me": "val 1",
                "me_also": "val 2",
                "unchanged": "val 3",
                1: "val 4",
            },
            False,
        ),
        (
            {
                "convertMe": "val 1",
                "meAlso": {
                    "nestedConvert": 1,
                    "unchanged": 2,
                    "anotherNest": {"oneMore": "hi"},
                },
                1: {"anotherNested": "a"},
            },
            {
                "convert_me": "val 1",
                "me_also": {
                    "nested_convert": 1,
                    "unchanged": 2,
                    "another_nest": {"one_more": "hi"},
                },
                1: {"another_nested": "a"},
            },
            False,
        ),
        (
            {"testConvert": [{"changeMe": 1, "unchanged": 2}, 1]},
            {"test_convert": [{"change_me": 1, "unchanged": 2}, 1]},
            False,
        ),
        (
            {"testConvert": ({"changeMe": 1, "unchanged": 2}, 1)},
            {"test_convert": ({"change_me": 1, "unchanged": 2}, 1)},
            False,
        ),
        (
            {"aTestWith12Number": 1, "AnotherTestWith12Number": 2, "endNumber1": 3},
            {
                "a_test_with_1_2_number": 1,
                "another_test_with_1_2_number": 2,
                "end_number_1": 3,
            },
            True,
        ),
    ],
)
def test_dict_to_snake(test_dict, expected, treat_digits_as_capitals):
    assert dict_to_snake(test_dict, treat_digits_as_capitals=treat_digits_as_capitals) == expected


@pytest.mark.parametrize(
    "test_str, expected_str",
    [("this_is_a_test", "thisIsATest"), ("this_is_a_12_test", "thisIsA12Test")],
)
def test_to_camel(test_str, expected_str):
    assert to_camel(test_str) == expected_str


@pytest.mark.parametrize(
    "test_str, expected_str",
    [("this_is_a_test", "ThisIsATest"), ("this_is_a_12_test", "ThisIsA12Test")],
)
def test_to_pascal(test_str, expected_str):
    assert to_pascal(test_str) == expected_str


@pytest.mark.parametrize(
    "test_str, expected_str, treat_digits_as_capitals",
    [
        ("thisIsATest", "this_is_a_test", False),
        ("ThisIsATest", "this_is_a_test", False),
        ("aTestWith12Number", "a_test_with12_number", False),
        ("aTestWith12Number", "a_test_with_1_2_number", True),
        ("ATestWith12Number", "a_test_with_1_2_number", True),
    ],
)
def test_to_snake(test_str, expected_str, treat_digits_as_capitals):
    assert to_snake(test_str, treat_digits_as_capitals=treat_digits_as_capitals) == expected_str


def test_class_to_camel():
    class SomeClass(Converter):
        some_value = "some value"

        def __init__(self) -> None:
            self.another_value = "another value"

    tester = SomeClass()

    assert tester.to_camel() == {"someValue": "some value", "anotherValue": "another value"}


def test_data_class_to_camel():
    @dataclass
    class SomeClass(Converter):
        some_value = "some value"

        def __init__(self) -> None:
            self.another_value = "another value"

    tester = SomeClass()

    assert tester.to_camel() == {"someValue": "some value", "anotherValue": "another value"}


def test_class_from_snake():
    class SomeClass(Converter):
        def __init__(self, some_value: str) -> None:
            self.some_value = some_value

    data = {"someValue": "some_value"}
    tester = SomeClass.from_camel(data)

    assert tester.some_value == data["someValue"]



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Any

from camel_converter._version import VERSION

__version__ = VERSION

if TYPE_CHECKING:
    import sys

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self


class Converter:
    """Inheriting this class adds methods to convert class and instance variables."""

    def to_camel(self) -> dict[str, Any]:
        """Converts to the class and instance variables into a camel case dict.

        Returns:
            A dictionary with they class and instance variable namess converted from snake case to
            camel case.

        Examples
            >>> from camel_converrter import Converter
            >>> class SomeClass(Converter):
            >>>     some_value = "some value"
            >>>
            >>>     def __init__(self) -> None:
            >>>         self.another_value = "another value"
            >>>
            >>> example = SomeClass()
            >>> print(example)

            This will print `{"someValue": "some value", "anotherValue": "another value"}`
        """
        class_vars = {
            key: value
            for key, value in self.__class__.__dict__.items()
            if not key.startswith("_") and not callable(value)
        }

        return dict_to_camel({**class_vars, **self.__dict__})

    @classmethod
    def from_camel(cls, data: dict[str, Any]) -> Self:
        """Initialize a class with a dictionary of camel case variables.

        Args:
            data: The dictionary to convert

        Returns:
            An instance of the class

        Examples
            >>> from camel_converrter import Converter
            >>> class SomeClass(Converter):
            >>>     def __init__(self, some_value: str) -> None:
            >>>         self.some_value = "some value"
            >>>
            >>> example = SomeClass.from_camel({"someValue": "some value"})
            >>> print(example.some_value)

            This will print `some value`
        """
        return cls(**dict_to_snake(data))


def dict_to_camel(data: dict[Any, Any]) -> dict[Any, Any]:
    """Converts dictionary keys from snake case to camel case.

    Only keys of type string are converted, any other type is left unchanged.

    Args:
        data: The dictionary to convert

    Returns:
        A dictionary with they keys of type string converted from snake case to camel case
    """
    converted: dict[Any, Any] = {}
    for k, v in data.items():
        if isinstance(k, str):
            key = to_camel(k)
        else:
            key = k

        if isinstance(v, dict):
            converted[key] = dict_to_camel(v)
        elif isinstance(v, list):
            converted[key] = [dict_to_camel(x) if isinstance(x, dict) else x for x in v]
        elif isinstance(v, tuple):
            converted[key] = tuple(dict_to_camel(x) if isinstance(x, dict) else x for x in v)
        else:
            converted[key] = v

    return converted


def dict_to_snake(
    data: dict[Any, Any], *, treat_digits_as_capitals: bool = False
) -> dict[Any, Any]:
    """Converts dictionary keys from camel case or pascal case to snake case.

    Only keys of type string are converted, any other type is left unchanged.

    Args:
        data:
            The dictionary to convert
        treat_digits_as_capitals:
            Whether to treat digits as capitals. For example myVariable2 would become my_variable_2
            rather than my_variable2.

    Returns:
        A dictionary with they keys of type string converted from camel case or pascal case to
        snake case
    """
    converted: dict[Any, Any] = {}
    for k, v in data.items():
        if isinstance(k, str):
            key = to_snake(k, treat_digits_as_capitals=treat_digits_as_capitals)
        else:
            key = k

        if isinstance(v, dict):
            converted[key] = dict_to_snake(v, treat_digits_as_capitals=treat_digits_as_capitals)
        elif isinstance(v, list):
            converted[key] = [
                dict_to_snake(x, treat_digits_as_capitals=treat_digits_as_capitals)
                if isinstance(x, dict)
                else x
                for x in v
            ]
        elif isinstance(v, tuple):
            converted[key] = tuple(
                dict_to_snake(x, treat_digits_as_capitals=treat_digits_as_capitals)
                if isinstance(x, dict)
                else x
                for x in v
            )
        else:
            converted[key] = v

    return converted


def dict_to_pascal(data: dict[Any, Any]) -> dict[Any, Any]:
    """Converts dictionary keys from snake case to pascal case.

    Only keys of type string are converted, any other type is left unchanged.

    Args:
        data: The dictionary to convert

    Returns:
        A dictionary with they keys of type string converted from snake case to pascal case
    """
    converted: dict[Any, Any] = {}
    for k, v in data.items():
        if isinstance(k, str):
            key = to_pascal(k)
        else:
            key = k

        if isinstance(v, dict):
            converted[key] = dict_to_pascal(v)
        elif isinstance(v, list):
            converted[key] = [dict_to_pascal(x) if isinstance(x, dict) else x for x in v]
        elif isinstance(v, tuple):
            converted[key] = tuple(dict_to_pascal(x) if isinstance(x, dict) else x for x in v)
        else:
            converted[key] = v

    return converted


@lru_cache(maxsize=4096)
def to_camel(snake_string: str) -> str:
    """Converts a snake case string to camel case.

    Args:
        snake_string: String in snake case format. For example my_variable.

    Returns:
        The string in camel case format. For example myVariable.
    """
    words = _split_snake(snake_string)

    return words[0] + "".join(word.title() for word in words[1:])


@lru_cache(maxsize=4096)
def to_snake(camel_string: str, *, treat_digits_as_capitals: bool = False) -> str:
    """Converts a camel case or pascal case string to snake case.

    Args:
        camel_string:
            String in camel case or pascal case format. For example myVariable.
        treat_digits_as_capitals:
            Whether to treat digits as capitals. For example myVariable2 would become my_variable_2
            rather than my_variable2.

    Returns:
        The string in snake case format. For example my_variable.
    """

    def check_character(c: str) -> str:
        if c.isupper() or (treat_digits_as_capitals and c.isdigit()):
            return f"_{c}"
        else:
            return c

    return "".join(check_character(c) for c in camel_string).lstrip("_").lower()


@lru_cache(maxsize=4096)
def to_pascal(snake_string: str) -> str:
    """Converts a snake case string to pascal case.

    Args:
        snake_string: String in snake case format. For example my_variable.

    Returns:
        The string in pascal case format. For example MyVariable.
    """
    words = _split_snake(snake_string)

    return "".join(word.title() for word in words)


def _split_snake(snake_string: str) -> list[str]:
    return snake_string.split("_")



import sys
from concurrent.futures import ThreadPoolExecutor
from importlib import reload
from unittest.mock import patch

import pytest

from camel_converter.pydantic_base import CamelBase


def test_camel_config():
    class Test(CamelBase):
        test_field: str

    got = Test(**{"testField": "test"})

    assert got.model_dump() == {"test_field": "test"}


def test_camel_config_missing_import():
    with patch.dict(sys.modules, {"pydantic": None}):
        with pytest.raises(ImportError):
            reload(sys.modules["camel_converter.pydantic_base"])


def test_camel_config_threaded():
    if not hasattr(sys, "_is_gil_enabled") or sys._is_gil_enabled():
        pytest.skip("Free-threading is not enabled")

    class Test(CamelBase):
        test_field: str

    def populate(values: dict) -> Test:
        return Test(**values)

    with ThreadPoolExecutor(max_workers=2) as executor:
        future1 = executor.submit(populate, {"testField": "value 1"})
        future2 = executor.submit(populate, {"testField": "value 2"})

        result1 = future1.result()
        result2 = future2.result()

    assert result1.model_dump() == {"test_field": "value 1"}
    assert result2.model_dump() == {"test_field": "value 2"}
